{**********************************************************}
{** UNIT    : Sonic                                      **}
{** LANGUAGE: Turbo Pascal 7.0                           **}
{** AUTHOR  : Marco Grubert; 100120,501                  **}
{** VERSION : 1.02                                       **}
{** OPTIONS : German                                     **}
{**********************************************************}
{** DESCRIPTION :                                        **}
{** Turbo Pascal Unit for playing Sonic Boom (.SB) files **}
{** on Sound Blaster compatible cards.                   **}
{** This is just an interface unit. The real work is done**}
{** by Tobias Downer's (74364,471) Sonic Driver...       **}
{**********************************************************}
{** This unit is public domaine. You may do with these   **}
{** files, as long as Tobias and I get the appropiate    **}
{** credits (please don't forget it !)                   **}
{** I am not responsible for any damages which may result**}
{** from the use of this unit or the demo program !      **}
{**********************************************************}
{** A word on the card configuration :                   **}
{** At startup this unit tries to read the blaster       **}
{** environment string to find out the correct values for**}
{** Address, Inter and DMA. If the environment string is **}
{** not present, default values (220h,7,1) will be used. **}
{** You may overwrite the detected values by chaning the **}
{** variables before (!) calling LoadDriver. Usually this**}
{** won't be necessary...                                **}
{**********************************************************}

{$X+}
UNIT Sonic;
INTERFACE



{$IFNDEF German}
CONST  ErrorMsg : ARRAY [-3..16] OF STRING = { English Messages }
      ('Sonic Driver Not Found'    ,
       'Not Enough Memory'         ,
       'Function Call Failed'      ,
       'Function Call Successful'  ,
       'Sample Already In Slot'    ,
       'Function Needs More Memory',
       'Soundcard Not Ready'       ,
       'No Soundcard Found'        ,
       'No Song Loaded'            ,
       'No BLASTER Environment'    ,
       'Driver Already Initialized',
       'DMA error'                 ,
       'Driver Not Initialized'    ,
       'Song Already Loaded'       ,
       'No Sample Loaded'          ,
       'Wrong Channel Number'      ,
       'Wrong Slot Number'         ,
       'Wrong Function Number'     ,
       'File I/O Error'            ,
       'Invalid .SB File'          );
{$ELSE}
CONST  ErrorMsg : ARRAY [-3..16] OF STRING = { Now German Messages }
      ('Sonic Treiber nicht gefunden'    ,
       'Nicht genug Speicher'            ,
       'Funktionsaufruf gescheitert'     ,
       'Functionsaufruf erfolgreich'     ,
       'Sample bereits in diesem Slot'   ,
       'Funktion bentigt mehr Speicher' ,
       'Soundkarte nicht bereit'         ,
       'Keine Soundkarte gefunden'       ,
       'Kein Song geladen'               ,
       'Keine BLASTER Umgebungsvariable' ,
       'Treiber bereits initialisiert'   ,
       'DMA Fehler'                      ,
       'Treiber nicht initialisiert'     ,
       'Song bereits geladen'      ,
       'Kein Sample geladen'       ,
       'Falsche Kanal Nummer'      ,
       'Falsche Slot Nummer'       ,
       'Falsche Funktionsnummer'   ,
       'Datei Ein/Ausgabefehler'   ,
       'Ungltige .SB Datei'       );
{$ENDIF}

VAR DriverError : INTEGER;     { Last error }

    Address      ,             { SB card information }
    Inter        ,             { If you want to override }
    DMA          : WORD;       { defaults then change these variables }
                               { before calling LoadDriver            }


{ Loads and initializes driver.                                 }
{ Path : Path to driver ( e.g.: 'C:\DOS\', '..\DRIVERS\', '' )  }
FUNCTION LoadDriver    (Path       : STRING)  : BOOLEAN;
{ If Driver was closed, reinitialize it                         }
FUNCTION InitDriver                           : BOOLEAN;
{ Close Driver temporary (speeds up program)                    }
FUNCTION CloseDriver                          : BOOLEAN;

{ Loads song into memory.                                            }
{ Filename: Complete filename (e.g.: 'POPCORN.SB','\SB\SONIC\TEST.SB }
FUNCTION LoadSong      (FileName   : STRING)  : BOOLEAN;
{ Frees memory after song has been played                            }
FUNCTION RemoveSong                           : BOOLEAN;

{ Plays loaded song.                                                 }
{ RepeatSong : If True the song will repeat until stopped by program }
FUNCTION PlaySong      (RepeatSong : BOOLEAN) : BOOLEAN;
{ Stops song currently being played                                  }
FUNCTION StopSong                             : BOOLEAN;

{ Load a sound sample into a memory region                           }
{ Filename : Complete filename (e.g.: 'D:\SB\TRAKS\Piano.SND')       }
{ Slot     : Memory region between 0 and 60                          }
{ ATTENTION : When loading a song, some slots are automatically being}
{             filled with samples, so use high numbers !             }
FUNCTION LoadSample    (FileName   : STRING;
                        Slot       : BYTE   ) : BOOLEAN;
{ Frees memory after sample has been played                          }
FUNCTION RemoveSample  (Slot       : BYTE   ) : BOOLEAN;
{ Plays a loaded sample.                                             }
{ Slot   : Memory region in which sample is saved (s.a.)             }
{ Volume : Volume for left and right channel                         }
{ Channel: Audio Channel (0-7 used for music / 8-15 for samples)     }
{ Frequ  : Playing frequency between 1 and 65535                     }
FUNCTION PlaySample    (Slot       ,
                        LeftVolume ,
                        RightVolume,
                        Channel    : BYTE   ;
                        Frequency  : WORD   ) : BOOLEAN;

{ Returns Volume for left and right channel                          }
FUNCTION GetVolume (VAR LeftVolume ,
                        RightVolume: BYTE   ) : BOOLEAN;
{ Set volume for left and right channel                              }
FUNCTION Volume        (Left       ,
                        Right      : BYTE   ) : BOOLEAN;

{ The following functions work when a song is being played. They     }
{ return the current position within the song (Module,Row,Pattern)   }
{ Use them to control program, e.g. switch pictures when a specific  }
{ point is reached...                                                }
FUNCTION GetPattern                           : WORD;
FUNCTION GetModule                            : WORD;
FUNCTION GetRow                               : WORD;

{ Returns an English / German error message of the last Error }
FUNCTION ErrorMessage                         : STRING;

IMPLEMENTATION

USES Dos;

{ Records the driver status }
CONST  Installed      : BOOLEAN = FALSE;

VAR
    DriverPointer: POINTER;  { Points to the loaded driver }
    DriverSize   : WORD   ;  { Size of the Sonic driver    }
    OldExit      : POINTER;  { Old Error Handler           }


{*****************************+ Internal Functions ***********************+}

{ This Function calls the Driver and passes various registers to it        }
FUNCTION ExecDriver(VAR Regs: Registers) : INTEGER;
VAR DriverPtr              : POINTER;
BEGIN
  ASM
    PUSH DS                         ;
    MOV Ax,WORD PTR DriverPointer   ; {Data Segment changes so move pointer}
    MOV Bx,WORD PTR DriverPointer+2 ; {on stack}
    MOV WORD PTR DriverPtr, Ax      ;
    MOV WORD PTR DriverPtr+2, Bx    ;
    LES DI, Regs                    ;
    PUSH ES                         ;
    PUSH DI                         ;
    MOV ah, Registers(ES:[DI]).&ah  ; {Copy Register variables to registers}
    MOV al, Registers(ES:[DI]).&al  ;
    MOV ch, Registers(ES:[DI]).&ch  ;
    MOV cl, Registers(ES:[DI]).&cl  ;
    MOV bx, Registers(ES:[DI]).&bx  ;
    MOV dx, Registers(ES:[DI]).&dx  ;
    MOV si, Registers(ES:[DI]).&si  ;
    MOV ds, Registers(ES:[DI]).&ds  ;

    CALL  DriverPtr                 ;   { call driver and pass registers}

    POP DI                          ;
    POP ES                          ;
    POP DS                          ;
    MOV Registers(ES:[DI]).&AX, AX  ; { save return value (for easier access}
    MOV DriverError, AX             ; { save error code }
    MOV @Result, AX                 ; { put return code on stack }
  END;
END;

{ This Procedure frees up the Driver's memory if it was loaded }
{$F+}
PROCEDURE UnInstallDriver;
VAR REgs : REgisters;
BEGIN
  ExitProc:=OldExit;
  IF DriverSize<>0 THEN
      FreeMem(DriverPointer,DriverSize);
  IF Installed THEN BEGIN
    CloseDriver;
  END;
END;
{$F-}

PROCEDURE ReadEnvironment;
VAR
    c        : INTEGER;
    Part     : STRING[5];
    EnvString: STRING;
BEGIN
  EnvString:=GetEnv('BLASTER');
  Address:=$220;
  Inter  :=7;
  DMA    :=1;
  IF EnvString<>'' THEN
    IF (Pos('A',EnvString[1])>0) AND (Pos('I',EnvString[1])>0) AND
    (Pos('D',EnvString[1])>0) THEN
    BEGIN
      Part:='$'+Copy(EnvString,Pos('A',EnvString)+1,3);
      Val(Part,Address,c);
      Part:='$'+Copy(EnvString,Pos('I',EnvString)+1,1);
      Val(Part,Inter,c);
      Part:='$'+Copy(EnvString,Pos('D',EnvString)+1,1);
      Val(Part,DMA,c);
    END;
END;

{************************** External Functions ****************************}

FUNCTION ErrorMessage: STRING;
BEGIN
  ErrorMessage:=ErrorMsg[DriverError];
END;

FUNCTION LoadDriver (Path : STRING): BOOLEAN;
VAR Filename : STRING;
    f        : FILE;
    Regs     : Registers;

BEGIN
  LoadDriver:=FALSE;
  Filename:=Path+'SONIC.DRV';
  Assign(f,Filename);
  {$I-}
  Reset(f,1);
  {$I+}
  IF IOResult<>0 THEN
  BEGIN
    DriverError:=-3;
    Exit;
  END;

  DriverSize:=Filesize(f);

  IF MaxAvail<DriverSize THEN
  BEGIN
    DriverError:=-2;
    DriverSize:=Maxavail;
    Exit;
  END;

  GetMem(driverpointer,DriverSize);

  BlockRead(F, driverpointer^,DriverSize);

  Close(f);

  { Now initializing Driver }

  IF InitDriver THEN LoadDriver:=TRUE;
END;

FUNCTION InitDriver : BOOLEAN;
VAR Regs : Registers;
BEGIN
  WITH Regs DO BEGIN
    Ah:=0;
    Al:=0;
    Bx:=Address;
    CH:=Inter;
    CL:=DMA;
  END;
  IF NOT Installed THEN BEGIN
    InitDriver:= ExecDriver(regs) = 0;
    Installed:=TRUE;
  END ELSE
    InitDriver:=FALSE;
END;

FUNCTION LoadSong(FileName:STRING):BOOLEAN;
VAR Regs: Registers;
    fn  : STRING;
BEGIN

  FN:=FileName+#0;

  WITH Regs DO BEGIN
    ah:=1;
    ds:=Seg(FN[1]);
    dx:=Ofs(FN[1]);
  END;

  LoadSong:= ExecDriver(Regs) = 0;
END;


FUNCTION RemoveSong : BOOLEAN;
VAR Regs: Registers;
BEGIN
  Regs.AH:=2;
  RemoveSong:= ExecDriver(Regs) = 0;
END;

FUNCTION PlaySong(RepeatSong:BOOLEAN):BOOLEAN;
VAR Regs: Registers;
BEGIN

  WITH Regs DO BEGIN
    ah:=3;
    al:=0; IF RepeatSong THEN al:=1;
    bx:=1;
  END;

  PlaySong:= ExecDriver(Regs) = 0;
END;

FUNCTION StopSong : BOOLEAN;
VAR Regs: Registers;
BEGIN
  Regs.AH:=4;
  StopSong:= ExecDriver(Regs) = 0;
END;

FUNCTION CloseDriver : BOOLEAN;
VAR Regs: REgisters;
BEGIN
  Regs.Ah:=5;
  IF Installed THEN
    CloseDriver:= ExecDriver(regs) = 0
  ELSE
    CloseDriver:=FALSE;
END;

FUNCTION LoadSample (FileName:STRING;Slot:BYTE):BOOLEAN;
VAR Regs: Registers;
    fn  : STRING;
BEGIN
  fn:=Filename+#0;
  WITH Regs DO BEGIN
    ah:=6;
    bx:=Slot;
    ds:=Seg(Fn[1]);
    dx:=Ofs(Fn[1]);
  END;

  LoadSample:= ExecDriver(Regs) = 0;
END;

FUNCTION RemoveSample(Slot:BYTE) : BOOLEAN;
VAR Regs: Registers;
BEGIN
  Regs.AH:=7;
  Regs.BX:=Slot;
  RemoveSample:= ExecDriver(Regs) = 0;
END;

FUNCTION PlaySample(Slot       ,
                    LeftVolume ,
                    RightVolume,
                    Channel    : BYTE;
                    Frequency  : WORD ) : BOOLEAN;
VAR Regs: Registers;
BEGIN
  WITH Regs DO BEGIN
    ah:=8;
    bx:=Slot;
    ch:=LeftVolume;
    cl:=rightVOlume;
    Dx:=Channel;
    Si:=Frequency;
  END;
  PlaySample:= ExecDriver(Regs) = 0;
END;

FUNCTION GetVolume(VAR LeftVolume ,
                       RightVolume: BYTE): BOOLEAN;
VAR Regs: Registers;
BEGIN
  Regs.Ah:=10;
  ExecDriver(Regs);

  GetVolume:=TRUE;
  LeftVolume :=Regs.AL;
  RightVolume:=Regs.AH;
END;

FUNCTION Volume (Left,Right:BYTE):BOOLEAN;
VAR Regs: Registers;
BEGIN

  WITH Regs DO BEGIN
    ah:=11;
    ch:=Left;
    cl:=Right;
  END;

  Volume  := ExecDriver(Regs) = 0;
END;


FUNCTION GetPattern: WORD;
VAR Regs: Registers;
BEGIN
  Regs.Ah:=12;
  ExecDriver(Regs);
  GetPattern:=Regs.AX;
END;

FUNCTION GetModule: WORD;
VAR Regs: Registers;
BEGIN
  Regs.Ah:=13;
  ExecDriver(Regs);
  GetModule:=Regs.AX;
END;

FUNCTION GetRow: WORD;
VAR Regs: Registers;
BEGIN
  Regs.Ah:=14;
  ExecDriver(Regs);
  GetRow:=Regs.AX;
END;


{ Here the ErrorHandler is replaced with my own one in order to free }
{ up the memory at the end of the program                            }
{ Also, the environment variable BLASTER is analyzed.                }
BEGIN
  DriverSize:=0;
  OldExit:=ExitProc;
  ExitProc:=@UnInstallDriver;
  ReadEnvironment;
END.